/**
 * filename: dcjdDist.h
 * author: Martin Bader
 * begin: 20.12.2007
 * last change: 04.08.2008
 *
 * Calculates the DJC + Duplications distance between two genomes.
 * 
 * Copyright (C) 2007  Martin Bader
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 
#ifndef DCJDDIST_H_
#define DCJDDIST_H_

#include <stack>
#include <utility>
#include <vector>

namespace dcjdDist
{

/**
 * An Operation.
 */
struct Operation
{
    /** Type of the operation. Note that we're working inverse, i.e.
     * when applying a duplication, we loose the duplicated markers */
    enum
    {
        OPERATION_INVALID,       // invalid operation
        OPERATION_DUPLICATION,   // duplication; for duplications, the first occurency is marked
        OPERATION_REVERSAL,      // reversal (DCJ with proper rejoining)
        OPERATION_DCJCIRC,       // a DCJ that builds a circular intermediate
        OPERATION_BLOCKINTERCHANGE,   // block interchange (2 DCJ) 
        OPERATION_DELETION       // deletion of a segment
    } m_type;
    
    /** Edges of the operation. Edges are always before the elements. */
    int m_edges[4];
    
    /** The deleted sequence (only used for deletions) */
    std::vector<int> m_sequence;
};



/**
 * The main class.
 */
class DcjdDist
{
    /***********************************************************************
     * member variables
     ***********************************************************************/
  public:
            
  private:
    /** Number of different markers */
    int m_numMarkers;
    
    /** Size of the actual permutation */
    int m_permSize;
    
    /** The permutation to sort (towards id) */
    int* m_perm;       
    
    /** the reality desire diagram: degree of each node */
    int* m_rddDegree; 
    
    /** the reality desire diagram: edges (as adjacency lists) */
    int** m_rddEdges;
      
    // variables needed for finding bridges (don't wanna reinitialize too 
    // often)
    /** timestamp of the nodes */
    int* m_timestamp;
    
    /** number of arriving backlinks of a node */
    int* m_numArrivingBacklinks;
    
    /** list of arriving backlinks */
    int** m_arrivingBacklinks;
    
    /** number of open backlinks passing the edge to parent */
    int* m_numOpenBacklinks;
    
    /** longest backlink for each node (source) passing this node */
    int* m_backlinkSource;
    
    /** longest backlink for each node (target) passing this node */
    int* m_backlinkTarget;
    
    /** target of longest backlink starting from this node */
    int* m_longestBacklink; 
    
    /** pointer to the interval stack of a node */
    std::stack<int>** m_intervalStacks;
    
    /** pointers to the corresponding storage points */
    std::stack<int>** m_intervalStores;
    
    /** indicates if an interval starting at a node is still open */
    bool* m_openInterval;
     
    /** mapping timestamp -> node */
    int* m_timestamp2Node;
    
    /** status of the node in the bridge finding algo:
     * 0: not in a bridge, or bridge reported elsewhere
     * 1: node in a onebridge
     * >= 2: node in a twobridge (with x edges in the 2-bridge cycle) */
    int* m_bridgeStatus;
    
    /** Bridges are reported here */
    std::pair<int, int>** m_bridges;
     
    /** Maps the bridges to the different components. */
    int* m_bridgeCompID;
    
    /** The actual component ID. */
    int m_compID; 
    
    /** Timestamp of deepest child */
    int* m_maxChild;
    
    /** This vector contains all good moves. */
    std::vector<Operation*> m_goodMoves;
    
    /** The scores of the moves */
    std::vector<int> m_scores;
    
    /** The scores of the sequences */
    std::vector<int*> m_seqscores;
    
    /** Small sequences that bring us closer to the id */
    std::vector<std::vector<Operation*> > m_sequences;
    
    /** Counts the selfloops of a node */
    int* m_numSelfloops;
    
    /** Temporary permutation (buffer for performing operations). */
    int* m_tmpPerm;
    
    /** For output: stack of operations of the sorting sequence. */
    std::stack<Operation*> m_sortingSequence;
    
    /** For output: sizes of the genomes in the sorting sequence. */
    std::stack<int> m_sortingSizes;
    
    /** For output: stack of all genomes on the sorting sequence. */
    std::stack<int*> m_sortingPerms;
                  
    /***********************************************************************
     * methods
     ***********************************************************************/     
  public:
    /**
     * Constructor.
     */
    DcjdDist();
        
    /**
     * Destructor.
     */
    ~DcjdDist();
    
    /** 
     * Finds a sorting sequence, and prints it to stdout.
     * @param f_numMarkers  num of different markers (= size of id)
     * @param f_permSize  size of the input genome
     * @param f_genome  the input genome
     */
    void sort(int f_numMarkers, int f_permSize, int* f_genome);     
    
  private:
    /**
     * Initializes the data strucutres required to find bridges 
     */
    void initBridges();
    
    /**
     * Finds all bridges of degree 1 and 2 in the RDD.
     */
    void findBridges();
    
    /** 
     * Expands a node during the depth first search for bridges.
     * The return value is the next empty timestamp
     * @param f_node  node index to expand
     * @param f_time  timestamp for this node
     * @param f_parent  index of parent
     * @return  a not yet used timestamp
     */
    int findBridgesExpand(int f_node, int f_time, int f_parent);    
    
    /**
     * Clears members used during findBridges.
     */
    void clearBridges();
    
    /** 
     * Lists all good moves (duplications with delta_c = 1, DCJ with 
     * delta_c = 2). The return value is the number of good moves.
     * Before calling this, findBridges must have been called.
     * @return  the number of good moves
     */
    int listGoodMoves();
    
    /**
     * Searches for all follow-ups such that the resulting operation is a 
     * valid blockinterchange. The operations will be added to m_goodMoves.
     * @param f_op  the operation to find the follow-ups
     * @return  the number of follow-ups
     */
    int getAllFollowUps(Operation& f_op);
    
    /**
     * Finds all possible reverse duplications, stores the corresponding 
     * reversals in m_goodMoves.
     */
    void findReverseDuplications();
    
    /**
     * Advanced heuristic to find useful deletions.
     * @param f_pos  position to break and insert
     * @return  pointer to the resulting operation
     */
    Operation* findDeletion3(int f_pos);
     
    /**
     * Scans for desire edges where an adjacency can be created. The 
     * sequences will be stored in m_sequences.
     * @return the number of found sequences
     */
    int findAdjacencySequences();
    
    /**
     * Scans for desire edges where both nodes have a degree <= 3, and creates 
     * sequences of operations that bring us closer to the id based on these 
     * edges. These found sequences will be stored in m_sequences.
     * @return the number of found sequences
     */
    int findType2Sequences();
    
    /** 
     * Scans for segments that occur at least 3 times in the permutation, and
     * creates sequences of operations that delete one of these segments 
     * without increasing the score. These found sequences will be stored in 
     * m_sequences.
     * @return  the number of found sequences
     */
    int findType3Sequences();
     
    /**
     * Finds transposition duplications. That are duplicated segements with 
     * the same orientation (they may be tandem duplications). The found 
     * sequences will be stored in m_sequences.
     * @return  the number of found sequences
     */
    int findTranspositionDuplications(); 
      
    /** 
     * All additional heuristics go here. The found sequences will be stored
     * in m_sequences.
     * @return  the number of found sequences
     */
    int findSpecialSequences();
       
    /**
     * Calculates the score of an operation, which is defined as follows:
     * 2 * delta lb / numOps, returns this as return value. Additionally 
     * calculates the more advanced score that is defined as follows:
     * 1) 60 * delta lb / seqweight
     * 2) -seqweight
     * 3) -2 * #missing elements + #duplicated elements
     * 4) #adjacencies
     * This result is written to f_score, which must be initialized as int[4].
     * @param f_op  the operation
     * @param f_lb  the old lower bound
     * @return  the score
     */
    int getScore(Operation& f_op, int f_lb, int* f_score);
     
    /**
     * Calculates the score of a sequence of operations, which is defined
     * as follows:
     * 1) 60 * delta lb / seqweight
     * 2) -seqweight
     * 3) -2 * #missing elements - #duplicated elements
     * 4) #adjacencies
     * The result is written to f_score, which must be initialized as int[4].
     * @param f_seq  the operation
     * @param f_lb  the old lower bound
     * @return  the score
     */
    void getSequenceScore(std::vector<Operation*>& f_seq, int f_lb, int* f_score);
     
    /**
     * Compares the scores of sequences of operations, which are stored as int[4] 
     * arrays. First compares the first entry, on equality the second, and so on.
     * Returns -1 if f_a < f_b, 0 if f_a == f_b, and +1 if f_a > f_b.
     * @param f_a  first score
     * @param f_b  second score
     * @return -1 if f_a < f_b, 0 if f_a == f_b, and +1 if f_a > f_b
     */
    int compSeqScores(int* f_a, int* f_b);
     
    /**
     * Performs an operation, and updates the RDD.
     * @param f_op  the operation to perform
     */
    void performOperation(Operation& f_op);
    
    /**
     * Returns the lower bound, which can be calculated as follows:
     * lb = n - 1 - C - sum_C_i (ceiling (s_i / 2))
     * where n is the size of the augmented permutation, C the number
     * of components, and s_i the number of nodes qith selfloop in
     * component C_i.
     * The lower bound is always calculated for the actual RDD.
     * @return  the lower bound
     */
    int getLowerBound();
    
    /***********************************************************************
     * debug methods
     ***********************************************************************/     
  public:
    /** 
     * Checks if the RDD and the permutation are consistent
     * @return  true if they are consistent
     */
    bool checkRDD(); 
     
    /**
     * Prints an operation to stdout.
     * @param f_op  the operation to print
     */
    void printOperation(Operation f_op);  

};


/***************************************************************************/

} // end namespace dcjdDist

#endif /*DCJDDIST_H_*/
